'---------------------------------------------------------------------------------------+
'                               TERMS OF USE: MIT License                               |
'                                                                                       |
' Permission is hereby granted, free of charge, to any person obtaining a copy of this  |
' software and associated documentation files (the "Software"), to deal in the Software |
' without restriction, including without limitation the rights to use, copy, modify,    |
' merge, publish, distribute, sublicense, and/or sell copies of the Software, and to    |
' permit persons to whom the Software is furnished to do so, subject to the following   |
' conditions:                                                                           |
'                                                                                       |
' The above copyright notice and this permission notice shall be included in all copies |
' or substantial portions of the Software.                                              |
'                                                                                       |
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,   |
' INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A         |
' PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT    |
' HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF  |
' CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE  |
' OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                                         |
'---------------------------------------------------------------------------------------+

             .section .ffds1.text, "ax"
              org 0

'----- Setup -----

             .global ffds1.entry
ffds1.entry   ' Counter A's job is to output data on the TX pin
              neg       PHSA, #1
              mov       CTRA, ffds1.ctra_val
              mov       DIRA, ffds1.maskTX            ' may be 0 if the TX pin is undefined

              ' Counter B's job is to count how many clocks the RX pin is low
              ' (specifically for aligning the read to the start bit).  Note:
              ' it increments by 2...I want to compare 1/4 the bit period, so
              ' instead of dividing the 1/2 bit period by 2, I just increment
              ' the logic counter by 2.
              mov       CTRB, ffds1.ctrb_val
              mov       FRQB, #2

              ' period info: load in 1/2 bit period
              rdlong    step_clocks, ffds1.period_ptr

              ' let the caller know it's OK to proceed (store a 0)
              wrlong    OUTB, PAR

              ' start looping
              mov       timestamp, step_clocks
              add       timestamp, CNT

'----- Receive -----
'
'       Notes:
'       * Any hub ops in RX must have a jmpret within 2 instructions: rdbyte nop jmpret
'

              ' resume after setting my start-bit-edge-detector to 0
rx_cleanup    mov       PHSB, #0

              ' loop (while calling TX co-routines) until we have a start bit
rx_main       jmpret    lockstep_ret, tx_jump
              mov       bits_in, PHSB wz
        if_z  jmp       #rx_main

              ' We have a start bit!
              ' Sync up with the correct place to sample the next bit:
              ' If RX was low for a short time (< 1/4 period), wait an
              ' extra 1/2 bit (places us before the last 1/4).
              jmpret    lockstep_ret, tx_jump
              cmp       bits_in, step_clocks wc,wz
        if_b  jmpret    lockstep_ret, tx_jump

              ' OK, we have 8 data bits left to read
              mov       bits_in, #8

              ' read in 8 bits
rx_main.get_bit      jmpret    lockstep_ret, tx_jump
              test      ffds1.maskRX, INA wc
              rcr       data_in, #1
              jmpret    lockstep_ret, tx_jump
              djnz      bits_in, #rx_main.get_bit

              ' align the data
              shr       data_in, #($20 - 8)

              ' read and verify the stop bit
              jmpret    lockstep_ret, tx_jump
              test      ffds1.maskRX, INA wc
        if_nc jmp       #rx_cleanup

              ' save the byte, increment the pointer
              add       rx_ptr, ffds1.rx_head_ptr
              wrbyte    data_in, rx_ptr
              add       rx_ptr, #1

              jmpret    lockstep_ret, tx_jump

              ' wrap the pointer and re-adjust back to 0-based index
              cmpsub    rx_ptr, ffds1.rx_end_ptr wc
        if_nc sub       rx_ptr, ffds1.rx_head_ptr

              ' update the head pointer
              wrword    rx_ptr, ffds1.update_head_ptr

              tjz       PHSB, #rx_main wr


'----- Sync -----

tx_jump       ' This is the jmpret register that handles switching to the right TX co-routine
              jmp       #tx_main

lockstep      ' This routine cycles through parts of the TX routine, and ends always locked in step
              waitcnt   timestamp, step_clocks
              ' send out another TX bit (if we have any) right after the lockstep
              ror       PHSA, write_bit
              ' return back to the RX code, also always a fixed time after the lockstep
lockstep_ret  jmp       #(0 - 0)


'----- Transmit -----
'
'       Notes:
'       * Any hub ops in TX must have a jmpret within 3 instructions: rdbyte nop nop jmpret
'       * There is no guarantee that a TX co-routine is in sync
'       * So bits are clocked out in "lockstep".
'

tx_main       ' check for outgoing data
              rdlong    bytes_out, PAR
              jmpret    tx_jump, #lockstep
              ' I need some free time to adjust the clock rate
              rdlong    step_clocks, ffds1.period_ptr
              jmpret    tx_jump, #lockstep
              ' loop if we have no data to send out
              tjz       bytes_out, #tx_main

              ' Bottom 16 bits hold the buffer address
              ' (no need to mask, prop does it for us)
              mov       write_ptr, bytes_out

              ' Top 16 bits hold the buffer length - 1
              shr       bytes_out, #16
              add       bytes_out, #1

              jmpret    tx_jump, #lockstep

tx_main.tx_byte ' set up for sending out a byte
              rdbyte    tmp, write_ptr
              add       write_ptr, #1

              ' force the stop bit
              or        tmp, #%1_00000000

              jmpret    tx_jump, #lockstep

              ' sign extend the 1 into all upper bits
              shl       tmp, #(32 - 9)
              sar       tmp, #(32 - 10)
              mov       PHSA, tmp

              ' 10 bits (start + 8 data + stop) makes 20 half-bits
              mov       half_bits_out, #20

tx_main.tx_bit ' wait 1/2 a bit, toggle the write_bit bit
              jmpret    tx_jump, #lockstep
              xor       write_bit, #1
              djnz      half_bits_out, #tx_main.tx_bit

              jmpret    tx_jump, #lockstep

              ' go back for more?
              djnz      bytes_out, #tx_main.tx_byte
              wrlong    bytes_out, PAR
              jmpret    tx_jump, #lockstep
              jmp       #(tx_main + 1)


'----- ASM parameters -----


             .global ffds1.period_ptr
             .global ffds1.maskRX
             .global ffds1.maskTX
             .global ffds1.rx_head_ptr
             .global ffds1.rx_end_ptr
             .global ffds1.ctra_val
             .global ffds1.ctrb_val
             .global ffds1.update_head_ptr
ffds1.period_ptr        long 0
ffds1.maskRX            long 0
ffds1.maskTX            long 0
ffds1.rx_head_ptr       long 0
ffds1.rx_end_ptr        long 0
ffds1.ctra_val          long 0
ffds1.ctrb_val          long 0
ffds1.update_head_ptr   long 0
rx_ptr                  long 0
write_bit               long 0


'----- ASM local variables -----


tmp                     res 1
step_clocks             res 1
bits_in                 res 1
half_bits_out           res 1
bytes_out               res 1
data_in                 res 1
timestamp               res 1
write_ptr               res 1


' make sure we fit inside the 512-16 instruction / variable limit
                        fit $1f0

